/*
    @(#) $Id: AxCrypt.cpp,v 1.25 2002/12/01 12:35:14 svante Exp $

	AxCrypt - Compressing and Encrypting Wrapper and Application Launcher for Secure Local,
	Server or Web Storage of Document Files.

	Copyright (C) 2002 Svante Seleborg/Axon Data, All rights reserved.

	This program is free software; you can redistribute it and/or modify it under the terms
	of the GNU General Public License as published by the Free Software Foundation;
	either version 2 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along with this program;
	if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
	Boston, MA 02111-1307 USA

	The author may be reached at mailto:axcrypt@axondata.se and http://axcrypt.sourceforge.net
----
	AxCrypt.cpp						WinMain() and friends.

	E-mail							YYYY-MM-DD				Reason
	axcrypt@axondata.se 			2001					Initial
	"								2001-11-19				SHChangeNotify on install/uninstall
	"								2001-11-29				Create new process instead of thread on
															first instance. Added some CAssertions.
															Changed from .App to .Sys in some cases.
															Fix for apps that don't start new
															processes. WaitForClose() et al.
	"								2001-12-15				Beta 5 - Heap and KeyStore as statics,
															not pointers, to help with construction etc.
                                    2002-08-06              Rel 1.2

*/
/*
	Compatibility:	Win 95/98/ME/NT/2K/XP.
					Attempts have been made to make a CE-port feasible - but not trivial.
					There is also an attempt to make it Unicode portable - untested...
					There is no hope of porting from the Windows platform. Sorry,
						that will require a partial re-write. It would be cool with a Mac-version and
						a Unix/Linux-version... Any takers ;-) ?

	A design goal has been to make it useful for low-bandwidth direct access
	situation such as intranets and web-based project repositories, therefore
	files are always compressed before encryption.

	Another design goal has been to use Windows API-code as much as possible,
	except for the actual crypto parts, since I want to make that part visible
	for full peer-review. Usage of Windows API's is somewhat limited by the idea
	of Win95 compatibility, especially some useful shell API's...

	Security Design Goals:

		-	A Key Encrypting Key can be broken by exaustive key search with a known plaintext-ciphertext
			pair, assuming that one Data Encrypting Key has been broken.
		-	The local system key management is secure, provided that the system is not crashed at a point
			where the memory mapped heap-file actually resides on disk.
		-	Data is kept locally secure, excepting untimely crashes, with the caveat that we have no control
			of other applications caretaking of the data, i.e. temp-files and paging-files from Word etc.

	The extension .axx is registered to start AxCrypt which
	then does the following:

	- Tries cached keys or asks if necessary for a key
	- Finds the original file name in the wrapped file
	- Decrypts
	- Expands to the original file name
	- Restores original file-times
	- Shells out to start the appropriate application
	- Waits for the started app to terminate.
	- Saves and checks file-times for update, if newer:
	- Compress
	- Encrypts using the same key to the original file name
	- Exits

	We are also hooked into the right-click properties menu of the shell
	with the action "Encrypt and Wrap with AxCrypt", as well as "Decrypt and Unwrap from AxCrypt".

	Internally there is one primary instance that responds to requests from
	subsequent instances. This is primarily to give central protected access to cached
	keys. This should probably be done through a COM-interface in the future...

	Each initiated action in the master instance creates a separate thread.
	There is no visible window - but one is created, among other things to listen and respond
	to termination requests such as log-off or shutdown. When asked to terminate, we will first
	wait for all threads to terminate before exiting ourselves.

	Since Named Pipes are not supported on Win95 we use a file-mapping view to
	communicate between master instance and requests. A mutex object is used to serialize the
	the use of the shared memory. When a request for encryption or decryption is to be sent,
	the process gets the mutex, writes a Request Struct to the file, sends a signal, waits for signal
	and finally releases the mutex when the master instance has copied the request and initiated a
	new thread.

	We keep track of all new threads in a simple manner, mostly to ensure that if the master instance
	is asked to exit, we can wait for all threads to finish first. We also use it to keep track of
	the key-data used for each thread.

	The code below and in associated files tend to be executed in different contexts, so
	some clarification may help... This executable may be the first instance, or it may
	not... We call that first instance the "primary", the rest "secondary".
	
	The code is presumably thread-safe, guarded with critical sections... I hope.

	A guide to my, sometimes inconsistent, dialect of hungarian may be in order...

	g - Global
	m_ - Class member
	s_ - Static (module local)
	h - Handle to something
	p - Pointer to something
	a - Array
	r - Reference to something
	b - Boolean (to be phased out and replaced with...)
	f - Also boolean...
	v - Void
	o - Octets (i.e. byte, unsigned char)
	e - Enum value/variable
	c - Constant (occasionally)
	ut - User Type, i.e. Classes and Structs. Sometimes not used.
	T - Type - Denotes that a name is the name of a 'type'
	C - Class - Denotes that a name is the name of a class
	sz - Zero-byte terminated string - or wrapper class for strings.
	dw - Double Word, i.e. unsigned 32-bit
	qw - Quad Word, i.e. unsigned 64-bit
	dqw - Double Quad Word, i.e. unsigned 128-bit
	ui - unsigned int
	i - int (as prefix, i.e. relatively small value that may take negative values)
	i,j etc - As simple names - Loop counters
*/
#include	"StdAfx.h"
//
#include	"stdlib.h"
#include	"shlobj.h"
#include    "CWinTrix.h"
#include    "CRegistry.h"
//
// This is to actually define the GUID in AxCryptGUID.h.
//
#define		INITGUID
#include	<initguid.h>
#include	"AxCryptGUID.h"
//
#include	"CActiveThreads.h"
#include	"GetOption.h"
#include	"CCryptoKey.h"
#include	"CFileName.h"
#include	"FileCmd.h"
#include	"CChildProc.h"
#include	"CCryptoHeap.h"
#include	"CVersion.h"
#include	"Dialog.h"
#include    "CEntropy.h"
#include	"CFileTemp.h"
//
// Command Line Parameters
//
extern int __argc;
extern char ** __argv;
//
// Type/Struct/Class declarations
//
struct SRequest {
	// In to primary
	enum eRequestType eRequest;
	TCHAR szParam1[MAX_PATH];       // First command parameter, if any.
    TCHAR szParam2[MAX_PATH];       // Second, ditto.
	TCHAR szIdTag[100];				// An id-tag, it must have some max-length...
    TCHAR szCurDir[MAX_PATH+1];     // The requesters current directory
                                    // Space for extra backslash too...
    BOOL fIgnoreEncrypted;          // Ignore already encrypted files.
    DWORD CallerProcId;
    DWORD dwBatch;                  // The batch id of this request.
    HWND hCurWnd;                   // The foreground window at the time of call
	HANDLE hStdOut;					// Valid in the primary context.

	// Out from primary to secondary
	HANDLE hWorkerThread;		
	DWORD dwExitCode;
    CProgressDialog *pDlgProgress;
	DWORD dwPrimaryProcessId;		// Primary process Id
};
//
// Static variables
//
static mfIsPrimary = false;

static HANDLE ghMutex = NULL, ghSendEvent = NULL, ghReceiveEvent = NULL;
//static CProgressDialog dlgProgress;
//
// Used to ensure handling of new thread list
//
static CActiveThreads* gpCActiveThreadsRoot = NULL;
static CRITICAL_SECTION gThreadListCritical;
//
// Does not need critical section protection, as the mutex is used
// for synchronized access to this structure.
//
static SRequest* glpSRequest = NULL;// The command buffer structure (memory mapped file view)
//
//	Globals
//
HWND ghWnd;							// The main window handle
HINSTANCE ghInstance;				// The main instance handle
//HWND ghProgressWnd;
//
//	Must be early, so as to be destructed late.
//
CHModule ghMsgModule;				// Need for messages
//
//	Define the secure heap unless running with C++ debug heap
//
#ifndef	_DEBUGHEAP
static BOOL gfHeapValid = FALSE;
CPtrTo<CCryptoHeap> pgHeap;
#endif	_DEBUGHEAP
//
//	Define the global key store root - our holiest of holies...
//
//	These must come after gHeap declaration, as they must constructed
//	in the order heap, then keystore and destructed in the reverse order.
//	C++ specifies that they are constructed in the order of declaration,
//	and destructed in the reverse order, ergo...
//
CPtrTo<CKeyList> pgKeyList;				// Our own "in-memory" key store.
//
//  Now define the PRNG and the Entropy Pool, *after* the global heap
//
CPtrTo<CEntropy> pgEntropyPool;

CPtrTo<CCryptoRand> pgPRNG;
//
#ifndef	_DEBUGHEAP
//
//	We override the new and delete operators to ensure that all allocs are 'secure'.
//
//	Special care is taken to handle the case of allocations/deletions when the heap
//	is not valid.
//
//	#define _DEBUGHEAP to use C++ standard heap instead - good for debugging!
//
#ifdef	_DEBUG
__declspec(thread) unsigned int tguiAlloc = 0;
#endif

void* operator new(size_t stOctets) {
	if (gfHeapValid && pgHeap->heap != NULL) {

#ifdef	_DEBUG
		// Add size of the overhead block, and align to even multiple of such block.
		size_t size = stOctets + sizeof UNIT + sizeof UNIT - 1;
		tguiAlloc += size & ~(sizeof UNIT - 1);
#endif

		return pgHeap->Alloc(stOctets);
	} else {
		return malloc(stOctets);
	}
}

void operator delete(void* vpMem) {
	if (!gfHeapValid ||
		pgHeap->heap == NULL ||
		vpMem < pgHeap->heap ||
		vpMem >= pgHeap->heap + pgHeap->m_qwFileSize) {
		free(vpMem);
	} else {

#ifdef	_DEBUG
		UNIT *p = (UNIT *)((unsigned)vpMem - sizeof UNIT);
		tguiAlloc -= (p->size & ~USED);
#endif

		pgHeap->Free(vpMem);
	}
}
#endif	_DEBUGHEAP
//
//	When launching apps, we need to be alone for a while so we can find exactly the right
//	child process.
//
CRITICAL_SECTION gLaunchAppCritical;
//
//	Walk and clean the list of active threads, if possible.
//	Use timeout as given in the argument. No guarantee of
//	empty list - check after call.
//
static void
PurgeThreadList(DWORD dwTimeOut) {
	CCriticalSection utThreadListCritical(&gThreadListCritical);

	utThreadListCritical.Enter();		// Destructor will ensure we leave the critical section
	CActiveThreads *pActiveThread = gpCActiveThreadsRoot;
	while (pActiveThread != NULL) {
		HANDLE hThread = pActiveThread->Thread();
		utThreadListCritical.Leave();

		// Wait for the thread to finish or whatever.
		DWORD dwReturn = MessageWaitForSingleObject(hThread, dwTimeOut);

		// If it is abandoned, remove it and reset the search pointer to the root.
		if (dwReturn == WAIT_OBJECT_0) {
			// This critical section stuff should really be done by CActiveThreads...
			utThreadListCritical.Enter();
			gpCActiveThreadsRoot->Remove(gpCActiveThreadsRoot, hThread);
			pActiveThread = gpCActiveThreadsRoot;
        } else {
		    utThreadListCritical.Enter();
		    pActiveThread = pActiveThread->Next();
        }
	}
	utThreadListCritical.Leave();
}
//
//	Clean up, ensure that threads are closed etc. It risks being called
//	twice, but should only attempt to do any work the first time.
//
//	When this function is called, the decision to exit must be final, and
//	memory allocations/deletions should not be called unless you know what
//	you are doing. We also close the mutex, so we really, really, should
//	exit when this is called.
//
//	exit() should be called immediately after calling this function.
//
static void
PrimaryPrepareForExit() {
	static fBeenHereOnce = FALSE;
	if (!fBeenHereOnce) {
		fBeenHereOnce = TRUE;

		pgEntropyPool->Stop();
		PurgeThreadList(1000);			// Give each thread 1 second
		(void)CloseHandle(ghMutex);
        ghMutex = NULL;
	}
}
//
// The main instance windows message process procedure
//
// "Primary Window Proc Main Thread"
//
static LRESULT CALLBACK
PrimaryWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    switch (uMsg) { 
        case WM_CREATE: 
            // Initialize the window. 
            break; 
 
        case WM_PAINT:
            // Paint the window's client area - actually do nothing to it.
			RECT stRect;
			if (GetUpdateRect(hWnd, &stRect, FALSE)) {
				PAINTSTRUCT stPaint;
				HDC hHDC = BeginPaint(hWnd, &stPaint);
				(void)EndPaint(hWnd, &stPaint);
			}
            break; 
		
		// Obsolete - but we support for older NT
		case WM_POWER:
			switch (wParam) {
			case PWR_SUSPENDREQUEST:
				CMessage().Wrap(0).AppMsg(INF_DEBUG, _T("PWR_SUSPENDREQUEST")).LogEvent(2);
				pgEntropyPool->Stop().Save();
				return PWR_OK;

			case PWR_CRITICALRESUME:
			case PWR_SUSPENDRESUME:
				pgEntropyPool->Start();
				return 0;
			}
			break;	// dummy...

		case WM_POWERBROADCAST:
			switch (wParam) {
			case PBT_APMSUSPEND:
				CMessage().Wrap(0).AppMsg(INF_DEBUG, _T("PBT_APMSUSPEND")).LogEvent(2);
				pgEntropyPool->Stop().Save();
				return 0;

			// Currently, always allow. In the future - check for ongoing transforms and
			// possible other cases that may conflict.
			case PBT_APMQUERYSUSPEND:
				return TRUE;

			// As we did not do anything during query suspend - we do not need to undo
			// anything here.
			case PBT_APMQUERYSUSPENDFAILED:
				return 0;

			case PBT_APMRESUMEAUTOMATIC:
			case PBT_APMRESUMECRITICAL:
			case PBT_APMRESUMESUSPEND:
				pgEntropyPool->Start();
				return 0;

			}
			return 0;	// Dummy
 
		case WM_QUERYENDSESSION:
			//
			// This is a case of difference between Windows NT/2K/XP etc on
			// the one hand and Windows 95/98/ME on the other.
			//
			// If there are no active apps, there is no problem, we always
			// allow shutdown.
			//
			// For Windows 95/98/ME:
			// We cannot allow the session to end with active sessions left,
			// since we will wind up waiting for other apps indefinitely,
			// unless we happen to be the last app to be asked. We cannot
			// depend on knowing, or arranging that order either. Thus we
			// inform the user of the fact, and refuse shutdown.
			//
			// For Windows NT/2K/XP etc:
			// If there are active apps, we ask if we should allow system shutdown,
			// as in this case the user can switch to the other apps and terminate
			// them before killing the dialog.
			//
			PurgeThreadList(0);		// Clean out any abandoned threads
			if (gpCActiveThreadsRoot == NULL) return TRUE;

            if (!COsVersion().IsWin9x()) {
				return CMessage().AppMsg(WRN_ACTIVE_APPS).ShowDialog(MB_OKCANCEL | MB_ICONWARNING) == IDOK;
			} else {
				CMessage().AppMsg(ERR_CANT_STOP).ShowError(MB_OK);
			}
			break;

		case WM_ENDSESSION:
			//
			// If WM_QUERYSESION allowed us to exit, wait for children to exit before...
			// It should be noted that when we get here, we cannot stop the shutdown, we
			// can only delay it.
			//
			if ((BOOL)wParam) {
				// If we are in NT-style os, we need to warn about active other programs.
				if (!COsVersion().IsWin9x()) {
					PurgeThreadList(0);		// Clean out any abandoned threads
					while (gpCActiveThreadsRoot != NULL) {
						CMessage().AppMsg(WRN_SHUT_DOWN).ShowWarning(MB_OK);
						PurgeThreadList(0);		// Clean out any abandoned threads
					}
				}
				PrimaryPrepareForExit();
				pgEntropyPool->Save();	// Always save entropy when asked to exit by system.
				// ...really terminating under our own control, ensuring clean-up
				exit(0);
			}
			// ... else just return 0 to say we was here
			break;

		case WM_DESTROY:
            PostQuitMessage(0);
			break;

        default: 
            return DefWindowProc(hWnd, uMsg, wParam, lParam); 
    } 
    return 0; 
}
//
// Register the Window Class
//
// "Primary Window Proc Main Thread"
//
static BOOL
PrimaryInitApplication(HINSTANCE hInstance) {
    WNDCLASSEX wcx;
	CVersion utVersion;
 
	ZeroMemory(&wcx, sizeof wcx);
    wcx.cbSize = sizeof wcx;			// size of structure 
    wcx.style = CS_HREDRAW | 
        CS_VREDRAW;						// redraw if size changes 
    wcx.lpfnWndProc = PrimaryWndProc;		// points to window procedure 
    wcx.cbClsExtra = 0;					// no extra class memory 
    wcx.cbWndExtra = 0;					// no extra window memory 
    wcx.hInstance = hInstance;			// handle of instance 
    wcx.hIcon = (HICON)LoadImage(
		hInstance, 
        MAKEINTRESOURCE(IDI_AXCRYPT),
		IMAGE_ICON,
		GetSystemMetrics(SM_CXICON),
		GetSystemMetrics(SM_CYICON),
		LR_DEFAULTCOLOR);				// Application Icon
	CAssert(wcx.hIcon != NULL).Sys().Throw();
	wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcx.hbrBackground = (HBRUSH)GetStockObject( 
        WHITE_BRUSH);					// white background brush 
    wcx.lpszClassName = utVersion.IntProductName();	// name of window class
    wcx.hIconSm = (HICON)LoadImage(hInstance,
        MAKEINTRESOURCE(IDI_AXCRYPT),
		IMAGE_ICON,
        GetSystemMetrics(SM_CXSMICON), 
        GetSystemMetrics(SM_CYSMICON), 
        LR_DEFAULTCOLOR);				// Small class icon
	CAssert(wcx.hIconSm != NULL).Sys().Throw();
 
    return RegisterClassEx(&wcx) != 0; 
} 
//
// Create a window to receive messages etc
//
// "Primary Window Proc Main Thread"
//
static BOOL
PrimaryInitInstance(HINSTANCE hInstance, int nCmdShow, HWND *phWnd) { 
	CVersion utVersion;
    *phWnd = CreateWindow( 
		utVersion.IntProductName(),// name of window class 
        utVersion.String(),		// title-bar string 
        WS_TILEDWINDOW,
        0, // Zero Size
        0,
        0,	// Zero Position
        0,
        (HWND)NULL,				// no owner window 
        (HMENU)NULL,			// use class menu 
        hInstance,				// handle of application instance 
        (LPVOID) NULL);			// no window-creation data 
 
    if (*phWnd == NULL) {
        return FALSE; 
	}
 
// Show the window and send a WM_PAINT message to the window 
// procedure. Hide it if we are in production.
#if		defined(_DEBUG)
    ShowWindow(*phWnd, nCmdShow);
#else
	ShowWindow(*phWnd, SW_HIDE);
#endif
    UpdateWindow(*phWnd); 
    return TRUE; 
}

struct FindProgressWndS {
    DWORD dwProcID;         // In
    HWND hProgressWnd;      // Out
};

static BOOL CALLBACK
EnumProgressProc(HWND hWnd, LPARAM lParam) {
    struct FindProgressWndS *pFindProgressWnd = (struct FindProgressWndS *)lParam;
    if (GetClassLong(hWnd, GCW_ATOM) == (DWORD)WC_DIALOG) {
        HWND hProgressWnd = FindWindowEx(hWnd, NULL, PROGRESS_CLASS, gszProgressWndName);
        if (hProgressWnd != NULL) {
            if ((DWORD)GetWindowLong(hProgressWnd, GWL_USERDATA) == pFindProgressWnd->dwProcID) {
                pFindProgressWnd->hProgressWnd = hProgressWnd;
                return FALSE;
            }
        }
    }
    return TRUE;    // Continue looking
}
//
//  Find a PROGRESS_CLASS window, that is a child window of a
//  top-level dialogue, with GWL_USERDATA == dwProcID.
//
static HWND
FindProgressWnd(DWORD dwProcID) {
    struct FindProgressWndS sFindProgressWnd;
    sFindProgressWnd.dwProcID = dwProcID;
    sFindProgressWnd.hProgressWnd = NULL;
    EnumWindows(EnumProgressProc, (LPARAM)&sFindProgressWnd);
    return sFindProgressWnd.hProgressWnd;
}
//
// This is where we do the actual work. As we want the message loop to continue
// we create a separate thread for it.
//
// The request buffer is allocated by the caller, and must be deallocated here
//
// "Primary Execute Request Thread", This is the actual worker thread.
//
static DWORD WINAPI
PrimaryCommandThread(LPVOID lpParam) {
	struct SRequest *lpSRequest = (SRequest *)lpParam;
	DWORD dwReturn = 0;

    // We used to need this, just keeping it as a neat trick
    // so I won't forget it... ;-)
    //
    // Force the creation of a message queue in this thread. This
	// is necessary for MessageBox() dialogs etc, even if we don't
	// actually have a message loop!
	//MSG msg;
	//PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);

    struct SCmdParam utCmdParam;
    try {
        ZeroMemory(&utCmdParam, sizeof utCmdParam);

        if (lpSRequest->szParam1[0]) utCmdParam.szParam1 = CopySz(lpSRequest->szParam1);
        if (lpSRequest->szParam2[0]) utCmdParam.szParam2 = CopySz(lpSRequest->szParam2);
		if (lpSRequest->szIdTag[0]) utCmdParam.szIdTag = CopySz(lpSRequest->szIdTag);

        // Pass the current directory as well.
        utCmdParam.szCurDir = CopySz(lpSRequest->szCurDir);

        utCmdParam.dwBatch = lpSRequest->dwBatch;
        utCmdParam.fIgnoreEncrypted = lpSRequest->fIgnoreEncrypted;
        utCmdParam.hForegroundWnd = lpSRequest->hCurWnd;
        utCmdParam.pDlgProgress = lpSRequest->pDlgProgress;
        utCmdParam.hProgressWnd = FindProgressWnd(lpSRequest->CallerProcId);
		utCmdParam.hStdOut = lpSRequest->hStdOut;
 
		HEAP_CHECK_BEGIN(_T("PrimaryCommandThread()"), 0)
		switch (lpSRequest->eRequest) {
		case EN_WIPE:   // Wipe
			dwReturn = CmdWipe(&utCmdParam);
			break;
		case EN_WIPES:  // Wipe silen
			dwReturn = CmdWipeSilent(&utCmdParam);
			break;
		case EN_OPEN:   // Launch and open, possibly re-encrypting
            // For the open command - we always use our own progress window
            if (lpSRequest->pDlgProgress) {
                if (utCmdParam.hProgressWnd = lpSRequest->pDlgProgress->Wnd()) {
                    // Set the file name in the progress window as well...
                    SetDlgItemText(GetParent(utCmdParam.hProgressWnd), IDS_FILE, utCmdParam.szParam1);
                }
            }
			dwReturn = CmdDecryptOpenLaunch(&utCmdParam);
            CWinTrix().ReallySetForegroundWindow(utCmdParam.hForegroundWnd);
            break;
        case EN_ENCRYPTZC: // Encrypt and Copy
            dwReturn = CmdEncryptZCFile(&utCmdParam, TRUE);
            break;
        case EN_ENCRYPTZCF: // Encrypt and Copy, Fast(er)
            // When in the 'fast' mode, we do not wipe temps.
            dwReturn = CmdEncryptZCFile(&utCmdParam, FALSE);
            break;
		case EN_ENCRYPTZ:   // Encrypt and wipe originals
			dwReturn = CmdEncryptZFile(&utCmdParam);
			break;
		case EN_DECRYPT:    // Decrypt and wipe originals.
			dwReturn = CmdDecryptFile(&utCmdParam);
			break;
		case EN_DECRYPTC:   // Decrypt and copy
			dwReturn = CmdDecryptCFile(&utCmdParam, TRUE);
			break;
		case EN_DECRYPTCF:  // Decrypt and copy, fast(er)
			dwReturn = CmdDecryptCFile(&utCmdParam, FALSE);
			break;
        case EN_ADDKEYENC:  // Add a default encryption key to the cache
            dwReturn = CmdAddKey(&utCmdParam, TRUE);
            break;
        case EN_ADDKEYDEC:  // Add a decryption key to the cache
            dwReturn = CmdAddKey(&utCmdParam, FALSE);
            break;
        case EN_ASKKEYENC:  // Ask for a (default) encryption key
            dwReturn = CmdPromptKey(&utCmdParam, TRUE);
            break;
        case EN_ASKKEYDEC:  // Ask for a decryption key
            dwReturn = CmdPromptKey(&utCmdParam, FALSE);
            break;
        case EN_CLEARKEYS:  // Clear all keys in cache, depending on batch id
            dwReturn = CmdClearKeys(&utCmdParam);
            break;
        case EN_RENAME:     // Rename to anonymous name.
            dwReturn = CmdAnonRename(&utCmdParam);
            break;
		case EN_SHOWTAG:	// Display the tag
			dwReturn = CmdShowIdTag(&utCmdParam);
			break;
		case EN_INSTALL:    // Install default settings in registry
			CmdInstallInRegistry(&utCmdParam);     // Just setup registry etc.
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST | SHCNF_FLUSHNOWAIT, 0, 0);
			break;
		case EN_UNINSTALL:  // Clear all registry entries
			CmdRemoveFromRegistry();	// Remove all settings from registry etc.
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST | SHCNF_FLUSHNOWAIT, 0, 0);
			break;
		case EN_PSPTEST:
			dwReturn = CChildProc().NeedPsapi();
			break;
        case EN_TESTHAVEKEY:    // Test if we have the key. Zero if we do...
            dwReturn = CmdTestHaveKey(&utCmdParam);
            break;
		case EN_EXIT:
			break;
        default:
            dwReturn = ERR_UNSPECIFIED;
            break;
		}
		HEAP_CHECK_END
		// Close handle to caller std out, if we have it.
		if (lpSRequest->hStdOut != NULL && lpSRequest->hStdOut != INVALID_HANDLE_VALUE) {
			CAssert(CloseHandle(lpSRequest->hStdOut)).Sys().Throw();
		}

		// We are no longer active - lets leave! ActiveThread's destructor also
		// closes the thread handle stored there.
		CCriticalSection utThreadListCritical(&gThreadListCritical);
		utThreadListCritical.Enter();
		gpCActiveThreadsRoot->Remove(gpCActiveThreadsRoot, lpSRequest->hWorkerThread);
		utThreadListCritical.Leave();
	} catch (TAssert utErr) {
		utErr.App(ERR_UNTRAPPED).Show();
        dwReturn = ERR_UNTRAPPED;
	}
    
    // Allocated by caller, must be released here
    if (lpSRequest->pDlgProgress) delete lpSRequest->pDlgProgress;
	delete lpSRequest;	

//#ifdef XXXYYY
    // In server-mode, we cause a special shell-out if the problem
    // was that a passphrase was needed!
    if (dwReturn == ERR_NO_PASSPHRASE) {
        CPtrTo<TCHAR> szCmdFmt = CRegistry(HKEY_CURRENT_USER, gszAxCryptRegKey, szRegValServerErrorShell).GetSz(_T(""));
        if (*szCmdFmt) {
            LPTSTR szCmd;
            CAssert(FormatMessage(FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_ARGUMENT_ARRAY,
                                  szCmdFmt,
                                  0,
                                  0,
                                  (LPTSTR)&szCmd,
                                  0,
                                  (va_list *)&utCmdParam.szParam1)).Sys(MSG_SYSTEM_CALL, _T("PrimaryCommandThread() [FormatMessage()]")).Throw();
        
            STARTUPINFO StartupInfo;
            ZeroMemory(&StartupInfo, sizeof StartupInfo);
            StartupInfo.cb = sizeof StartupInfo;

            PROCESS_INFORMATION ProcessInfo;
            if (CreateProcess(NULL,
                              szCmd,
                              NULL,
                              NULL,
                              FALSE,
                              0,
                              NULL,
                              NULL,
                              &StartupInfo,
                              &ProcessInfo)) {
                CMessage().Wrap(0).AppMsg(INF_SERVER_SHELL, szCmd).LogEvent(1);
            } else {
                DWORD dwErr = GetLastError();
                CMessage().Wrap(0).AppMsg(ERR_SERVER_SHELL, szCmd, CMessage().SysMsg(dwErr).GetMsg()).LogEvent(0);
            }
            LocalFree(szCmd);
        }
    }
//#endif
    return dwReturn;
}
//
//  Do all primary-specific initialization
//
static void
PrimaryInit(int nCmdShow) {
    // Initialize application and instance
    CAssert(PrimaryInitApplication(ghInstance)).Sys(MSG_INIT_APPLICATION).Throw();
    CAssert(PrimaryInitInstance(ghInstance, nCmdShow, &ghWnd)).Sys(MSG_INIT_INSTANCE).Throw();

    // Initialize the global entropy pool
    pgEntropyPool = new CEntropy(HKEY_CURRENT_USER, gszAxCryptRegKey);

    // Start with cleaning temp directory, if needed.
    // Force focus from original window - otherwise me may lock-up the temp directory
    // if that is viewed in windows explorer.
    MySetForegroundWindow();

    // Clean the old temp directory, if any. We need the entropy generator running,
    // as CTempDir needs random numbers.
    pgEntropyPool->Load().Start();
    CTempDir().SetPath2TempDir().RemoveDir();
    pgEntropyPool->Stop().Save();
            
    // Create new tempdirectory and ensure that it is not compressed
	// These calls and the need for it only exists in NT and later
    if (COsVersion().IsWinNx()) {
        CFileIO utTmpDir;
        utTmpDir.OpenDir(CFileName().SetPath2TempDir().GetDir());
        utTmpDir.SetNotCompressed();
        utTmpDir.Close();
    }

    // Then make the secure heap
    pgHeap = new CCryptoHeap(SECURE_HEAP_SIZE, &gfHeapValid);
    if (pgHeap == NULL) {
	    FatalAppExit(0, _T("Can't allocate secure heap - aborting."));
    }
    pgHeap->Init();

    // ...and the global key store
    pgKeyList = new CKeyList;
    if (pgKeyList == NULL) {
	    FatalAppExit(0, _T("Can't allocate key store - aborting."));
    }

    // ...and the PRNG
    pgPRNG = new CCryptoRand;
    if (pgPRNG == NULL) {
	    FatalAppExit(0, _T("Can't allocate PRNG - aborting."));
    }

    InitializeCriticalSection(&gThreadListCritical);
    InitializeCriticalSection(&gLaunchAppCritical);

    pgEntropyPool->Load().Start();

	CMessage().Wrap(0).AppMsg(INF_APP_START).LogEvent();
    CAssert(ReleaseMutex(ghMutex)).Sys(MSG_SYSTEM_CALL, _T("ReleaseMutex() [StartPrimaryProcess()]")).Throw();
}
//
//  Handle reception of primary process event request. Actually do some
//  work by spawning a worker thread for it.
//
static void
PrimaryEvent() {
    //
    //	At this point, we know we have received a request,
    //	the caller has the mutex, but has relinquished control
    //	of the shared memory to us through the event.
    //

	// First we check if this is a request for our processid.
	// If so, we do that here, and then just return.
	if (glpSRequest->eRequest == EN_GETPROCID) {
		glpSRequest->dwPrimaryProcessId = GetCurrentProcessId();
	    CAssert(SetEvent(ghReceiveEvent)).Sys(MSG_SYSTEM_CALL, _T("SetEvent()")).Throw();
		return;
	}

    // Get a copy of the request code.
    enum eRequestType eRequest = glpSRequest->eRequest;

    // Copy the command structure, start a new thread for the real action
    // and abandon it to it's own devices. Unless it's an exit event we
    // don't wait for it to finish.
    SRequest *lpSRequestCopy = new SRequest;
    CAssert(lpSRequestCopy != NULL).App(MSG_MEMORY_ALLOC, _T("PrimaryEvent()")).Throw();

    DWORD dwThreadID; // We don't use it, but win 9x requires it
    // The copy is delete'd by the worker thread, so it must not
    // be used after we release the thread from suspension!
    CopyMemory(lpSRequestCopy, glpSRequest, sizeof SRequest);

    // Deleted by the command thread.
    if (eRequest == EN_OPEN) {
        lpSRequestCopy->pDlgProgress = new CProgressDialog();
        lpSRequestCopy->pDlgProgress->Create(ghInstance, IDD_PROGRESS, glpSRequest->hCurWnd);
    } else {
        lpSRequestCopy->pDlgProgress = NULL;
    }

    // Start the worker thread suspended so we can add it to the active
    // list first. It cannot be done in that thread, as the process may
    // terminate before it actually starts execution. Tried that ;-(.
    CAssert((lpSRequestCopy->hWorkerThread = CreateThread(NULL,
	    0,
	    PrimaryCommandThread,
	    lpSRequestCopy,
	    CREATE_SUSPENDED,
	    &dwThreadID)) != NULL).Sys(MSG_SYSTEM_CALL, _T("CreateThread()")).Throw();

    // Give the caller the thread handle to wait for
    HANDLE hThreadWait = lpSRequestCopy->hWorkerThread;

    // Don't let the worker thread affect foreground too much if it's a long operation
    CAssert(SetThreadPriority(hThreadWait, THREAD_PRIORITY_BELOW_NORMAL)).Sys(MSG_SYSTEM_CALL, _T("PrimaryEvent() SetThreadPriority()")).Throw();
    switch (eRequest) {
    case EN_EXIT:
    case EN_UNINSTALL:
    case EN_PSPTEST:
    case EN_INSTALL:
        hThreadWait = GetCurrentThread();
        break;
    default:
	    break;
    }

    CHandle hCallerProc = OpenProcess(PROCESS_DUP_HANDLE, FALSE, glpSRequest->CallerProcId);
    CAssert(hCallerProc.IsValid()).Sys(MSG_SYSTEM_CALL, _T("OpenProcess() [WinMain()]")).Throw();
    CAssert(
	    DuplicateHandle(
		    GetCurrentProcess(),
		    hThreadWait,
		    hCallerProc,
		    &glpSRequest->hWorkerThread,
		    SYNCHRONIZE|THREAD_QUERY_INFORMATION,
		    FALSE,
		    0)).Sys(MSG_SYSTEM_CALL, _T("DuplicateHandle()")).Throw();
    CAssert(hCallerProc.Close()).Sys(MSG_SYSTEM_CALL, _T("CloseHandle(hCallerProc) [WinMain()]")).Throw();

    // We can release the interprocess buffer, and let the caller get on with it.
    CAssert(SetEvent(ghReceiveEvent)).Sys(MSG_SYSTEM_CALL, _T("SetEvent()")).Throw();

    // Add the new thread to the active list.
    CCriticalSection utThreadListCritical(&gThreadListCritical);
    utThreadListCritical.Enter();
    new CActiveThreads(gpCActiveThreadsRoot, lpSRequestCopy->hWorkerThread);
    utThreadListCritical.Leave();

    // Make a copy of the handle for processing below, after we resume the
    // worker thread, we cannot use lpSRequestCopy
    CHandle hWorkerThread;
    CAssert(DuplicateHandle(GetCurrentProcess(), lpSRequestCopy->hWorkerThread, GetCurrentProcess(), &hWorkerThread, 0, FALSE, DUPLICATE_SAME_ACCESS)).Sys(MSG_SYSTEM_CALL, _T("WinMain [DuplicateHandle]")).Throw();

    // Now we can start the thread to actually run. lpSRequestCopy is subsequently invalid!!!
    CAssert(ResumeThread(lpSRequestCopy->hWorkerThread) != 0xffffffff).Sys(MSG_SYSTEM_CALL, _T("ResumeThread() [WinMain()]")).Throw();

    // Save the pool for all but some request-types.
    switch (eRequest) {
    case EN_UNINSTALL:
    case EN_PSPTEST:
    case EN_INSTALL:
	    break;
    default:
	    // Event, Save the Pool on every event except the above
	    pgEntropyPool->Save();
    }

    // Terminate main instance for some request-types
    switch (eRequest) {
    case EN_EXIT:
    case EN_UNINSTALL:
    case EN_PSPTEST:
    case EN_INSTALL:
	    CAssert(MessageWaitForSingleObject(hWorkerThread, 5000) != WAIT_FAILED).Sys(MSG_SYSTEM_CALL, _T("WinMain [WaitForSingleOjbect(hWorkerThread)]")).Throw();
	    
	    // Pass on the exit/uninstall/psptest thread return code as the process return code
	    DWORD dwExitCode;
	    CAssert(GetExitCodeThread(hWorkerThread, &dwExitCode)).Sys(MSG_SYSTEM_CALL, _T("WinMain [GetExitCodeThread]")).Throw();
	    PostQuitMessage(dwExitCode);

    default:
	    break;
    }
    // Close the work thread handle always.
    CAssert(hWorkerThread.Close()).Sys(MSG_SYSTEM_CALL, _T("WinMain [hWorkerThread.Close()]")).Throw();
}
//
//  Do all necessary work to end the primary process.
//
static void
PrimaryEnd() {
    // First let's give active threads up to a second to finish.
    for (int i = 0; i < 20; i++) {
	    if (gpCActiveThreadsRoot == NULL) break;
	    PurgeThreadList(0);
	    Sleep(50);	// Wait just a little for potentially active threads.
    }
    // While we have any potentially active threads,
    // let's warn the user about this, and ...
    while (gpCActiveThreadsRoot != NULL) {
    //							FlashBox();
	    CMessage().AppMsg(WRN_ACTIVE_APPS).ShowWarning(MB_OK);
    }
    // ... then clean up etc.
    PrimaryPrepareForExit();
}
//
//
//
void
DestroyGlobals() {
    // Delete the PRNG..
    pgPRNG = NULL;

    // .. and delete the global key store...
    pgKeyList = NULL;

    // ... and and delete the secure heap...
    pgHeap = NULL;

    // .. and the entropy pool
    pgEntropyPool = NULL;
}
//
//
//
void CleanTmp() {
    if (mfIsPrimary) {
        CTempDir().SetPath2TempDir().RemoveDir();
        CMessage().Wrap(0).AppMsg(INF_APP_QUIT).LogEvent();
    }
}
//
//  Run the primary process - it's ok to try many times.
//  Only one is started though.
//
//  Try to create the Mutex, and get initial ownership
//
//  Return when we want to exit.
//
static int
PrimaryProcess(int nCmdShow) {
    // Create/Open the Mutex, and attempt to get initial ownership
    ghMutex = CreateMutex(NULL, TRUE, gszAxCryptMutex);
    // If the call succeeded...
    if (ghMutex) {
        // .. and if we actually created it
        if (GetLastError() != ERROR_ALREADY_EXISTS) {
            // ... then we are the one and only master. The only case when this
            // is not true, is if the master is killed, and we have secondaries
            // still alive. This can be fixed by introducing yet another Mutex
            // that is always owned by the master.
            PrimaryInit(nCmdShow);

            // We are now the master server process, waiting for requests for action
			// through the shared memory map - signalled through SendEvent semaphore.
			while (TRUE) {
				MSG msg ; 
				while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) { 
					if (msg.message == WM_QUIT) {
                        PrimaryEnd();
						return msg.wParam;
					} else {
						TranslateMessage(&msg); 
						DispatchMessage(&msg); 
					}
				}
				if (MsgWaitForMultipleObjects(1,
					&ghSendEvent, 
					FALSE,
					INFINITE,
					QS_ALLINPUT) == (WAIT_OBJECT_0)) {
                        PrimaryEvent();
#ifdef	_DEBUGPLUS
				        {
					        // Use this code to check heap alloc requirements
					        static size_t stMaxAlloc;
					        size_t stCurAlloc = gHeap.CurrentAlloc();
					        if (stCurAlloc > stMaxAlloc) {
						        stMaxAlloc = stCurAlloc;
						        TCHAR szAlloc[100];
						        sprintf(szAlloc, "New max alloc=%d", stMaxAlloc);
						        MessageBox(NULL, szAlloc, CVersion().String(), MB_OK);
					        }
				        }
#endif	_DEBUGPLUS
				}
            }
        }
    } else {
	    CAssert(FALSE).App(MSG_CREATE_MUTEX, (LPTSTR)gszAxCryptMutex).Throw();
    }
    return 0;
}
//
//	Call the main process for action
//
//	If return is FALSE; error occurred and *pdwExitCode is set.
//
//  Secondary Main Thread
//
static BOOL
SecondaryExecuteRequest(SRequest *pRequest, DWORD *pdwExitCode) {
    *pdwExitCode = 0;
    try {
		HANDLE hThread;

        // Get exclusive access to the request buffer.
        CAssert(MessageWaitForSingleObject(ghMutex) == WAIT_OBJECT_0).Sys().Throw();

        // Now get the primary process id.
		ZeroMemory(glpSRequest, sizeof *glpSRequest);
		glpSRequest->eRequest = EN_GETPROCID;
		CAssert(SetEvent(ghSendEvent)).Sys().Throw();
		CAssert(MessageWaitForSingleObject(ghReceiveEvent, INFINITE) == WAIT_OBJECT_0).Sys().Throw();
		
		// Now we know the primary process id. Lets open a handle to it.
		CHandle hPrimaryProcess = OpenProcess(PROCESS_DUP_HANDLE, FALSE, glpSRequest->dwPrimaryProcessId);
		CAssert(hPrimaryProcess != NULL).Sys().Throw();
		
		// Get the new request into the global request buffer.
        *glpSRequest = *pRequest;
		
		// Get handle to stdout into the the global request buffer, if possible
		//
		// It's not possible to duplicate console handles to other processes, and...
		// ...GUI process are not created with a console, and are not attaced to a
		// console even when started from the command prompt. Thus, we must check
		// that we have a valid stdout beore passing it.
		// The main process will close the handle passed to it, if any.
		HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
		if (hStdOut == INVALID_HANDLE_VALUE ||
			GetFileType(hStdOut) != FILE_TYPE_DISK &&
			GetFileType(hStdOut) != FILE_TYPE_PIPE) {

			glpSRequest->hStdOut = INVALID_HANDLE_VALUE;
		} else {
			CAssert(DuplicateHandle(GetCurrentProcess(), hStdOut, hPrimaryProcess, &glpSRequest->hStdOut, 0, FALSE, DUPLICATE_SAME_ACCESS)).Sys().Throw();
		}

        // Pass our current directory (reserve room for extra backslash)
        DWORD dwLen;
        CAssert((dwLen = GetCurrentDirectory(sizeof glpSRequest->szCurDir / sizeof glpSRequest->szCurDir[0] - 1, glpSRequest->szCurDir)) != 0)
            .Sys(MSG_SYSTEM_CALL, _T("SecondaryExecuteRequest() [GetCurrentDirectory()]"))
            .Throw();
        if (glpSRequest->szCurDir[dwLen - 1] != _T('\\')) {
            glpSRequest->szCurDir[dwLen++] = _T('\\');
            glpSRequest->szCurDir[dwLen] = _T('\0');
        }

		glpSRequest->CallerProcId = GetCurrentProcessId();
        
        // Disable foreground window handing in server mode
        if (CRegistry(HKEY_CURRENT_USER, gszAxCryptRegKey, szRegValServerMode).GetDword(FALSE)) {
            glpSRequest->hCurWnd = NULL;
        } else {
            glpSRequest->hCurWnd = GetForegroundWindow();
        }

		CAssert(SetEvent(ghSendEvent)).Sys().Throw();

		// We need to process messages in this thread as well
		// Wait for main thread to acknowledge reception of event and
		// start of new worker thread, if any.
		if (MessageWaitForSingleObject(ghReceiveEvent, INFINITE) != WAIT_OBJECT_0) {
			MessageBox(NULL, "SecondaryExecuteRequest [MessageWaitForObject]", "Wait failed", MB_OK);
		}

		hThread = glpSRequest->hWorkerThread;
		//
		// here we need to get hold of the thread id, wait for it after the release mutex
		// so that we stay in memory until all is done.
		//
		// Let someone else have a go... Release the mutex
		CAssert(ReleaseMutex(ghMutex)).Sys().Throw();
		if (hThread != NULL) {
			if (MessageWaitForSingleObject(hThread) == WAIT_FAILED) {
                CAssert(FALSE).Sys(MSG_SYSTEM_CALL, _T("SecondaryExecuteRequest [MessageWaitForSingleObject(Threads)]")).Throw();
				//MessageBox(NULL, "Wait failed", "", MB_OK);
			}
			CAssert(GetExitCodeThread(hThread, pdwExitCode)).Sys(MSG_SYSTEM_CALL, _T("SecondaryExecuteRequest [GetExitCodeThread]")).Throw();
			CloseHandle(hThread);
		}
	} catch (TAssert utErr) {
		(void)ReleaseMutex(ghMutex);
		utErr.File(MSG_CMD_LINE_OPEN, pRequest->szParam1).Show();
		*pdwExitCode = utErr.LastError();
	}
    return *pdwExitCode ? FALSE : TRUE;
}
//
// Parse command line for action - This is never done in the primary instance.
//
// -b nnnnn             Define the batch that this command is part of.
// -t                   Clear all keys (of a batch if given, otherwise all keys).
// -i					Install application
// -u					Uninstall application
// -x					Exit the application - mostly for testing purposes
// -p					Test for need to install psapi.dll
// -c filename(s)       Encrypt and Wrap, do not wipe original.
// -d filename(s)		Decrypt and Unwrap, do not open
// -w filename(s)       Wipe
// (-o) filename(s)		Open and Launch
//
// "Secondary Main Thread"
//
//
static  DWORD
SecondaryParseCommand() {
	DWORD dwExitCode = 0;
	try {
        SRequest Request;
        ZeroMemory(&Request, sizeof Request);

        enum eRequestType eRequest = EN_OPEN;

        LPTSTR szParam1 = NULL; TCHAR chOpt;
        LPTSTR szOutFileName = NULL, szIdTag = NULL;
        BOOL fAddKeyEncrypt = FALSE;
        BOOL fFast = FALSE, fCopy = FALSE;
    	while (chOpt = GetOption(__argc, __argv, "ab:cdefghi:k:n:opqstuwxy:Yz", &szParam1)) {
            if (szParam1) {
                lstrcpyn(Request.szParam1, szParam1, sizeof Request.szParam1 / sizeof Request.szParam1[0]);
            } else {
                Request.szParam1[0] = _T('\0');
            }
			switch (chOpt) {
            case 'a':	// Ask for key
                Request.eRequest = fAddKeyEncrypt ? EN_ASKKEYENC : EN_ASKKEYDEC;
                if (!SecondaryExecuteRequest(&Request, &dwExitCode)) {
                    return dwExitCode;
                }
                break;
            case 'b':	// Batch id.
                Request.dwBatch = atol(szParam1);
                break;
            case 'c':
                // Toggle copy-instead flag.
                fCopy = !fCopy;
                break;
			case 'd':
                // First check for disabled decryption mode
                if (gfNoDecryptMode) {
			        CMessage().AppMsg(MSG_UNKNOWN_OPT, _T("-d")).ShowError();
			        return MSG_UNKNOWN_OPT;
                }

                // Decrypt, depending on fast and copy flags.
                if (fCopy) {
                    if (fFast) {
                        eRequest = EN_DECRYPTCF;
                    } else {
                        eRequest = EN_DECRYPTC;
                    }
                } else {
				    eRequest = EN_DECRYPT;
                }
				break;
            case 'e':
                fAddKeyEncrypt = TRUE;
                break;
            case 'f':
                fFast = !fFast;
                break;
            case 'g':
                Request.fIgnoreEncrypted = !Request.fIgnoreEncrypted;
                break;
            case 'h':
                eRequest = EN_RENAME;
                break;
			case 'i':
                Request.eRequest = EN_INSTALL;
                SecondaryExecuteRequest(&Request, &dwExitCode);
				return dwExitCode;
            case 'k':	// Add key
                Request.eRequest = fAddKeyEncrypt ? EN_ADDKEYENC : EN_ADDKEYDEC;
                _tcsset(szParam1, _T('*'));

                if (!SecondaryExecuteRequest(&Request, &dwExitCode)) {
                    ZeroMemory(&Request, sizeof Request);
                    return dwExitCode;
                }
                ZeroMemory(Request.szParam1, sizeof Request.szParam1);
                break;
            case 'n':
                szOutFileName = szParam1;
                break;
			case 'o':
				eRequest = EN_OPEN;
				break;
			case 'p':
                Request.eRequest = EN_PSPTEST;
				SecondaryExecuteRequest(&Request, &dwExitCode);
				return dwExitCode;
            case 'q':
                eRequest = EN_TESTHAVEKEY;
                break;
			case 's':
				eRequest = EN_WIPES;
				break;
            case 't' :
                Request.eRequest = EN_CLEARKEYS;
                if (!SecondaryExecuteRequest(&Request, &dwExitCode)) {
                    return dwExitCode;
                }
                break;
			case 'u':
                Request.eRequest = EN_UNINSTALL;
				SecondaryExecuteRequest(&Request, &dwExitCode);
				return dwExitCode;
			case 'w':
				eRequest = EN_WIPE;
				break;
			case 'x':
                Request.eRequest = EN_EXIT;
				SecondaryExecuteRequest(&Request, &dwExitCode);
				return dwExitCode;
			case 'y':
				// IdTag.
				// Valid only on encryption, it cannot be set on encrypted files.
				// Potentially it can be implemented if the key is known.
				// It is retained unchanged on -o with re-encryption.
				szIdTag = szParam1;
				break;				
			case 'Y':	// Show tag if any of following encrypted files.
				eRequest = EN_SHOWTAG;
				break;
			case 'z':
                // Encrypt, depending on fast and copy flags.
                if (fCopy) {
                    if (fFast) {
                        eRequest = EN_ENCRYPTZCF;
                    } else {
                        eRequest = EN_ENCRYPTZC;
                    }
                } else {
				    eRequest = EN_ENCRYPTZ;
                }
				break;
			case 1:	// Execute currently valid request option
                // Set output filename parameter for those options that take one.
                switch (eRequest) {
                case EN_ENCRYPTZ:
                case EN_ENCRYPTZC:
                case EN_ENCRYPTZCF:
					if (szIdTag != NULL) {
	                    lstrcpyn(Request.szIdTag, szIdTag, sizeof Request.szIdTag / sizeof Request.szIdTag[0]);
					}
					// fall through...!
                case EN_DECRYPT:
                case EN_DECRYPTC:
                case EN_DECRYPTCF:
				case EN_SHOWTAG:
					if (szOutFileName != NULL) {
						lstrcpyn(Request.szParam2, szOutFileName, sizeof Request.szParam2 / sizeof Request.szParam2[0]);
						szOutFileName = NULL;
					}
                    break;
                default:
                    Request.szParam2[0] = _T('\0');
                }
                Request.eRequest = eRequest;
                if (!SecondaryExecuteRequest(&Request, &dwExitCode)) {
                    // We continue file processing, even if a request was just ignored.
                    if (dwExitCode != WRN_IGNORED) {
                        return dwExitCode;
                    }
                }
				break;
			default:
				CMessage().AppMsg(MSG_UNKNOWN_OPT, szParam1).ShowError();
				return MSG_UNKNOWN_OPT;
			}
            Request.szParam1[0] = _T('\0');
			szParam1 = NULL;
		}
	} catch (TAssert utErr) {
		utErr.App(MSG_PARSE_COMMAND_INTERNAL).Show();
		return utErr.LastError();
	}
	return dwExitCode;
}
//
// Start here - Split depending on first instance or not. The first
// instance sets itself up as a server, secondary instances call it.
//
//  If we have no first, or primary, instance we launch a new process
//  for it. The primary is loaded by calling with no parameters.
//  It's ok to try multiple times, in that case it's just a noop.
//
int APIENTRY
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    int iReturn = 0;
    HANDLE hRequestFileMap = NULL;

	// Set up for messages and logging
    InitGlobalStrings(ghInstance = hInstance);

    atexit(CleanTmp);
    atexit(DestroyGlobals);
    atexit(UnInitGlobalStrings);

    if (!(ghMsgModule = LoadLibraryEx(
		CFileName().SetPath2ExeName(hInstance).SetTitle((LPTSTR)gszAxCryptMessageDLL).Get(),
		NULL, 
		LOAD_LIBRARY_AS_DATAFILE)).IsValid()) {
		FatalAppExit(0, _T("Failed to load application texts. Immediate exit."));
	}

    try {
        // Common initialization, regardless of primary or secondary
		// We create/open two Events (one to send, one to acknowledge reception) and
		// a file mapping view to handle the request communication. If the file mapping already
		// exists, we assume that is because we are a secondary instance.
        CAssert((ghSendEvent = CreateEvent(NULL, FALSE, FALSE, gszAxCryptEventSend)) != NULL).App(MSG_CREATE_EVENT, gszAxCryptEventSend).Throw();
		CAssert((ghReceiveEvent = CreateEvent(NULL, FALSE, FALSE, gszAxCryptEventReceive)) != NULL).App(MSG_CREATE_EVENT, gszAxCryptEventReceive).Throw();
		CAssert((hRequestFileMap = CreateFileMapping((HANDLE)0xFFFFFFFF,
											 NULL,
											 PAGE_READWRITE,
											 0,
											 sizeof SRequest,
											 gszAxCryptFileMap)) != NULL).App(MSG_CREATE_REQUEST_MAP, gszAxCryptFileMap).Throw();
		CAssert((glpSRequest = (SRequest *)MapViewOfFile(hRequestFileMap,
			FILE_MAP_WRITE,
			0, 0,
			sizeof SRequest)) != NULL).App(MSG_CREATE_REQUEST_MAP, gszAxCryptFileMap).Throw();

        // If a commandline, then we parse command and call primary.
        if (_tcslen(lpCmdLine) > 0) {
            // Ensure that the primary is started
            ghMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, gszAxCryptMutex);
            // If we couldn't open the mutex, we need a primary process
            if (!ghMutex) {
				STARTUPINFO stStartupInfo;
				ZeroMemory(&stStartupInfo, sizeof stStartupInfo);
				stStartupInfo.cb = sizeof stStartupInfo;
				PROCESS_INFORMATION stProcessInformation;
				ZeroMemory(&stProcessInformation, sizeof stProcessInformation);
				
				CAssert(CreateProcess(
					NULL,
                    CFileName().SetPath2ExeName(hInstance).GetQuoted(),
					NULL,
					NULL,
					FALSE,
					0,
					NULL,
					NULL,
					&stStartupInfo,
					&stProcessInformation)).Sys(MSG_SYSTEM_CALL, _T("CreateProcess() [WinMain()]")).Throw();
                
                WaitForInputIdle(stProcessInformation.hProcess, INFINITE);
                CAssert(CloseHandle(stProcessInformation.hThread)).Sys(MSG_SYSTEM_CALL, _T("CloseHandle() [WinMain() .hThread]")).Throw();
                CAssert(CloseHandle(stProcessInformation.hProcess)).Sys(MSG_SYSTEM_CALL, _T("CloseHandle() [WinMain() .hProcess]")).Throw();
            }
            // Now we _must_ be able to open the mutex!
            CAssert((ghMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, gszAxCryptMutex)) != NULL).Sys(MSG_SYSTEM_CALL, _T("OpenMutex() [WinMain()]")).Throw();
            iReturn = SecondaryParseCommand();
        } else {
            mfIsPrimary = true;
            iReturn = PrimaryProcess(nCmdShow);
        }
    } catch (TAssert utErr) {
		utErr.App(MSG_WINMAIN).Show();
		iReturn = utErr.LastError();
	}
    // Close all open handles etc
    if (ghSendEvent) CloseHandle(ghSendEvent);
    if (ghReceiveEvent) CloseHandle(ghReceiveEvent);
    if (ghMutex) CloseHandle(ghMutex);
    if (glpSRequest) UnmapViewOfFile(glpSRequest);
    if (hRequestFileMap) CloseHandle(hRequestFileMap);
    return iReturn;
}
