/*
    @(#) $Id: CXform.cpp,v 1.9 2002/08/23 16:42:03 svante Exp $

	AxCrypt - Compressing and Encrypting Wrapper and Application Launcher for Secure Local,
	Server or Web Storage of Document Files.

	Copyright (C) 2001 Svante Seleborg/Axon Data, All rights reserved.

	This program is free software; you can redistribute it and/or modify it under the terms
	of the GNU General Public License as published by the Free Software Foundation;
	either version 2 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along with this program;
	if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
	Boston, MA 02111-1307 USA

	The author may be reached at mailto:axcrypt@axondata.se and http://axcrypt.sourceforge.net
----
	CXform.cpp						File transformation operations such as encrypt/decrypt/compress/decompress/wipe etc
	
	E-mail							YYYY-MM-DD				Reason
	axcrypt@axondata.se 			2001					Initial
									2001-11-27				Fixed loop in CWipe with partial wipe > 1K
                                    2002-08-02              Rev 1.2

*/
#include	"StdAfx.h"
#include	"CXform.h"
#include    "CVersion.h"
#include    "commctrl.h"
//
//	Initialize the stream structure. Mostly for ZLib purposes, but we use the data counters
//	for the other derived classes as well.
//
CXform::CXform(HWND hProgressWnd) {
	// Initialize the stream
	utZstream.avail_in = 0;
	utZstream.avail_out = 0;
	utZstream.total_in = 0;
	utZstream.total_out = 0;
	utZstream.zalloc = (alloc_func)0;
	utZstream.zfree = (free_func)0;
	utZstream.opaque = (voidpf)0;

    m_hProgressWnd = hProgressWnd;
}
//
//  Default init
//
/*virtual*/ QWORD
CXform::Init(CFileIO& utInFile) {
    return utInFile.m_qwFileSize - utInFile.m_qwFileOffset;
}
//
//	Generic data transformation using a stream to control in and out. It will
//	read and write to the stream, which ensure that all blocking and buffering
//	requirements are met.
//
//	This class job is to get and provide streams of bytes to the underlying
//	algorithms which include compression/decompress/encryption/decryption/hashing/wiping etc.
//
void
CXform::XformData(CFileIO& utInFile, CFileIO& utOutFile) {
	HEAP_CHECK_BEGIN(_T("CXform::XformData()"), 0);

	QWORD qwTotalInputSize = Init(utInFile);
	CAssert(qwTotalInputSize >= 0).File(MSG_FILE_SIZE_ERROR, utInFile.FileName()).Throw();
    
    // Set the name of the operation in the progress window, if any
    StartProgress();

    // Do the transformation, in multiple segments if necessary.
	while (utZstream.total_in < qwTotalInputSize) {
		// Go for gold - Try to map the entire input file. MapView will adjust
		// if necessary to keep within other limits. We must also stay below 32-bit signed limit.
		QWORD qwInputSize = qwTotalInputSize - utZstream.total_in;
		if (qwInputSize < 0x80000000) {
			utZstream.avail_in = (DWORD)qwInputSize;
		} else {
			utZstream.avail_in = 0x7fffffff;
		}
		utZstream.next_in = utInFile.MapView(
			FILE_MAP_READ,
			utZstream.total_in,
			&utZstream.avail_in);
		// Now consume all input to be compressed in this segment.
		while (utZstream.avail_in > 0) {
            // Update progress bar, use different strategies for cases where
            // times 100 risks an overflow, and when it does not.
            if (qwTotalInputSize < 0x80000000) {
                QWORD qwConsumedTimes100 = utZstream.total_in * 100;
                Progress((unsigned int)(qwConsumedTimes100 / qwTotalInputSize));
            } else {
                QWORD qwSizeDiv100 = qwTotalInputSize / 100;
                Progress((unsigned int)(utZstream.total_in / qwSizeDiv100));
            }
            EnsureOut(utOutFile);
			Xform();
		}
	}
	// Flush the final stuff. If necessary, make room for possible last squirt.
	do {
		EnsureOut(utOutFile);
	} while (Finish());
	End();
	utInFile.CloseMap();
	utOutFile.CloseMap();
	utOutFile.SetFilePointer(utOutFile.m_qwFileSize = utOutFile.m_qwFileOffset + utZstream.total_out);
	utOutFile.SetEndOfFile();
	HEAP_CHECK_END
	return;
}
//
//	Helper for my stream handlers that use the z_stream structure from ZLib.
//
void
CXform::ConsumedInOut(DWORD dwIn, DWORD dwOut) {
	utZstream.avail_in -= dwIn;
	utZstream.total_in += dwIn;
	utZstream.next_in += dwIn;

	utZstream.avail_out -= dwOut;
	utZstream.total_out += dwOut;
	utZstream.next_out += dwOut;
}
//
//	Default implementation is return to zero. The purpose of this function is to
//	handle the case when we do not beforehand know the exact size of the output
//	data. This currently only happens during compression - all other derived class
//	should refrain from redefining this function, as all necessary out-space will
//	be allocated by the regular code.
//
int
CXform::EstimateOut() {
	return 0;
}
//
// Update progress indicator.
//
void
CXform::Progress(unsigned int iPercent) {
    if (m_hProgressWnd != NULL) {
        if (!GetWindowLong(GetParent(m_hProgressWnd), GWL_USERDATA)) {
            CAssert(FALSE).App(WRN_CANCEL).Throw();
        }
        PostMessage(m_hProgressWnd, PBM_SETPOS, (WPARAM)(iPercent > 100 ? 100 : iPercent), 0);
    }
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CXform::StartProgress() {
    if (m_hProgressWnd != NULL) {
        // Clear the operation text. Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)_T(""));

        // Start the visual wait timer, use send message to ensure sequence.
        SendMessage(GetParent(m_hProgressWnd), WM_APP + 2, 0, 0);
    }
}

//
//	Ensure that avail_out > 0, or guarantee that that no output will need be written.
//	This may be overridden to do nothing for transformations that produce no output.
//
void
CXform::EnsureOut(CFileIO& utFile) {
	if (utZstream.avail_out == 0) {
		//
		// Here we have two basic cases:
		// 1 - We know the exact output total (all cases except compression).
		// 2 - We don't know how large the output will be (only compression).
		//
		// The tough part is optimizing case 2 - we do not want to allocate unnecessary
		// space on disk, but do not want to allocate in too small chunks either...
		// 
		// utFile.dwSize should be interpreted as being the maximum known size of
		// the output - as long as we don't exceed this with utZstream.total_out we are
		// all right. When it is about to be exceeded, we need a hint on how much to increase the
		// output file size. This is provided by the EstimateOut() function.
		//
		// That it is exceeded is seen by the fact that the entire reserved/known file
		// space (dwFileSize) is used up by the current transformation at this point.
		//
		// The call to EstimateOut() is an escape to the specific transformation-routine,
		// which normally returns 0, as most transformations are to known output sizes,
		// except compression.
		//
		QWORD qwAvailOut = utFile.m_qwFileSize - (utFile.m_qwFileOffset + utZstream.total_out);
		if (qwAvailOut < 0x80000000) {
			utZstream.avail_out = (DWORD)qwAvailOut;
		} else {
			utZstream.avail_out = 0x7fffffff;
		}
		// If no space, this means the outfile is of unknown size and needs growing, so we ask the
		// transformation (i.e. compression) for an estimate of needed allocation size.
		if (utZstream.avail_out == 0) {
			utZstream.avail_out = EstimateOut();
		}
		// If still zero, means the transform guarantees that no extra space is needed for Finish().
		if (utZstream.avail_out > 0) {
			utZstream.next_out = utFile.MapView(
				FILE_MAP_WRITE,
				utZstream.total_out,
				&utZstream.avail_out);
		}
	}
}
//
//	Wrap ZLib Init()
//
/*virtual*/ QWORD
CCompress::Init(CFileIO& utInFile) {
	CAssert(deflateInit(&utZstream, Z_DEFAULT_COMPRESSION) == Z_OK).App(MSG_DEFLATE_INIT).Throw();
    return utInFile.m_qwFileSize - utInFile.m_qwFileOffset;
}
//
//	Call deflate for each section of data from the input file as it appears.
//
/*virtual*/ void CCompress::Xform() {
	CAssert(deflate(&utZstream, Z_SYNC_FLUSH) == Z_OK).App(MSG_DEFLATE_SYNC).Throw();
}
//
//	Flush all buffers and write the last output to the output buffer
//
/*virtual*/ int
CCompress::Finish() {
	int iReturn = deflate(&utZstream, Z_FINISH);
	if (iReturn == Z_OK) return TRUE;

	CAssert(iReturn == Z_STREAM_END).App(MSG_COMPRESS_FINISH).Throw();
	return FALSE;
}
//
//	Free datastructures etc.
//
/*virtual*/ void
CCompress::End() {
	CAssert(deflateEnd(&utZstream) == Z_OK).App(MSG_COMPRESS_FINISH).Throw();
}
//
//	Estimate how much output space we will require. This governs how large chunks may be written, and
//	possibly how large the file may grow during deflation (it will be truncated at the end anyway),
//	but theoretically this could result in file-system-full exception if we estimate way to much space.
//
//	This is a gross approximation... We start by chancing a 50% ratio, this is just to try to not
//	allocate too large files unnecessarily when we do the mapping.
//	At the very, very end, there is a risk that all avail_in has been consumed, but we still need
//	additional space to flush the last buffered. Then we just give a small amount.
//
//	Also limit EstimatedOut-size to 32 bits signed - if the file is *much* larger, this may cause
//	some ineffiency, but we'll just have to live with that.
//
/*virtual*/ int
CCompress::EstimateOut() {
	if (utZstream.avail_in == 0) {
		//	This is kind of complicated, it turns out that the compressor can need a *lot* of space
		//	after avail_in == 0, this happens when no real compression occurs for example.
		//	So we try for some intelligent allocation strategy...
		//
		// If we are actually expanding data, give at least 100 bytes or 1% of in-data.
		if (utZstream.total_out >= utZstream.total_in) {
			return Max(100, (DWORD)Min(utZstream.total_in/100, 0x7fffffff));
		}
		// If we already have produced more than 75% of total_in, give to fill up to 100%
		if (utZstream.total_out >= (utZstream.total_in/2 + utZstream.total_in/4)) {
			return (DWORD)Min(utZstream.total_in - utZstream.total_out, 0x7fffffff);
		}
		// if more than 50% of input
		if (utZstream.total_out >= utZstream.total_in/2) {
			return (DWORD)Min(utZstream.total_in/2, 0x7fffffff);
		}
		// 1000 or 5% of total input size, whichever is the largest
		return (DWORD)Min(Max(1000, utZstream.total_in/20), 0x7fffffff);
	} else {
		if (utZstream.total_out == 0) {
			return utZstream.avail_in/2;		// First time.
		} else {
			return utZstream.avail_in;			// up-coming times.
		}
	}
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CCompress::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_COMPRESS).GetMsg());
    }
}
//
//	Wrap ZLib Init(), and limit reading of the file to the start for this purpose.
//
/*virtual*/ QWORD
CCompressRatio::Init(CFileIO& utInFile) {
	CAssert(deflateInit(&utZstream, Z_DEFAULT_COMPRESSION) == Z_OK).App(MSG_DEFLATE_INIT).Throw();
    if (CXform::Init(utInFile) < COMPRESS_TEST_SIZE) {
        return CXform::Init(utInFile);
    } else {
        return COMPRESS_TEST_SIZE;
    }
}
//
//	Call deflate for each section of data from the input file as it appears.
//
/*virtual*/ void
CCompressRatio::Xform() {
	CAssert(deflate(&utZstream, Z_SYNC_FLUSH) == Z_OK).App(MSG_DEFLATE_SYNC).Throw();
}
//
//	Ensure that avail_out > 0, or guarantee that that no output will need be written.
//
/*virtual*/ void
CCompressRatio::EnsureOut(CFileIO& utOutFile) {
    utZstream.next_out = m_aoDummyBuf;
    utZstream.avail_out = sizeof m_aoDummyBuf;
}
//
//	Flush all buffers and write the last output to the output buffer
//
/*virtual*/ int
CCompressRatio::Finish() {
	int iReturn = deflate(&utZstream, Z_FINISH);
	if (iReturn == Z_OK) return TRUE;

	CAssert(iReturn == Z_STREAM_END).App(MSG_COMPRESS_FINISH).Throw();
	return FALSE;
}
//
//	Free datastructures etc.
//
/*virtual*/ void
CCompressRatio::End() {
	CAssert(deflateEnd(&utZstream) == Z_OK).App(MSG_COMPRESS_FINISH).Throw();
    //
    // Some tricks to keep to integer arithmetic without risk of overflow etc,
    // calculate the ratio between the diference of out and in.
    // 100 is perfect compress, 0 is no compression.
    //
    QWORD qwTemp = utZstream.total_in / 100;
    if (qwTemp > 0) {
        m_iRatio =  (int)(100 - utZstream.total_out / qwTemp);
    } else if (utZstream.total_in > 0) {
        m_iRatio =  (int)(100 - utZstream.total_out * 100 / utZstream.total_in);
    } else {
        m_iRatio = 0;
    }

    if (m_iRatio < 0) {
        m_iRatio = 0;
    }

    if (m_iRatio > 100) {
        m_iRatio = 100;
    }
}
//
//  Return the calculated compression ratio.
//
int
CCompressRatio::GetRatio() {
    return m_iRatio;
}
//
//	Decompress transformer class begins here
//
//	Wrap ZLib inflateInit()
//	
/*virtual*/ QWORD
CDecompress::Init(CFileIO& utInFile) {
	utZstream.next_in = Z_NULL;	// Defer check to first call to inflate
	CAssert(inflateInit(&utZstream) == Z_OK).App(MSG_INFLATE_INIT).Throw();
    return utInFile.m_qwFileSize - utInFile.m_qwFileOffset;
}
//
//	Wrap ZLib call to inflate() for each segment as it arrives from the input
//
/*virtual*/ void
CDecompress::Xform() {
	int iReturn = inflate(&utZstream, Z_SYNC_FLUSH);
	if (iReturn == Z_OK || iReturn == Z_STREAM_END) return;
	CAssert(FALSE).App(MSG_INFLATE_ERROR).Throw();
}
//
//	Wrap ZLib call to flush its buffers to the output file
//
/*virtual*/ int
CDecompress::Finish() {
	int iReturn = inflate(&utZstream, Z_FINISH);
	if (iReturn == Z_OK) return TRUE;
	CAssert(iReturn == Z_STREAM_END).App(MSG_INFLATE_FINISH).Throw();
	return FALSE;
}
//
//	Wrap ZLib call to End() to clear it's data structures.
//
/*virtual*/ void
CDecompress::End() {
	CAssert(inflateEnd(&utZstream) == Z_OK).App(MSG_INFLATE_END).Throw();
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CDecompress::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_DECOMPRESS).GetMsg());
    }
}

/*virtual*/ QWORD
CWipe::Init(CFileIO& utInFile) {
	m_utRnd.Seed(NULL, 0);			// Seed with only internally generated entropy, for what it is worth.
    return utInFile.m_qwFileSize - utInFile.m_qwFileOffset;
}
//
//	Fill all available out-space with pseudo-random data.
//
/*virtual*/ void
CWipe::Xform() {
	// it is the in-file which determines how much to wipe, as we are doing a pseudo-transformation.
	// In some cases we do not want to wipe the entire out-file.
	DWORD dwLen = Min(utZstream.avail_in, utZstream.avail_out);
	m_utRnd.RandomFill(utZstream.next_out, dwLen);
	ConsumedInOut(dwLen, dwLen);
}
//
//	As we do not internally buffer anything - nothing need be done here.
//	FALSE return means "Don't call me again, I'm done."
//
/*virtual*/ int
CWipe::Finish() {
	return FALSE;
}
//
//	No cleaning up is necessary either.
//
/*virtual*/ void
CWipe::End() {
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CWipeTemp::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_WIPETEMP).GetMsg());
    }
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CWipeData::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_WIPEDATA).GetMsg());
    }
}
//
//	CBlockXform
//
//	Abstract base class for block-oriented transformer classes
//	begins here.
//	It helps implement transformations that needs input in fixed
//	block sizes, and buffers data appropriately regardless of how
//	the byte stream arrives.
//	Specifically encryption/decryption, but it may be used for other
//	cases as well.
//
CBlockXform::CBlockXform(HWND hProgressWnd, DWORD dwBlkSiz) : CXform(hProgressWnd) {
	// Initialize the stream temp block
	m_utBlkBuf.dwBlkSiz = dwBlkSiz;				// Set-up block size
	m_utBlkBuf.poIn = new BYTE[m_utBlkBuf.dwBlkSiz];	// need to be secure, as they may contain cleartext.
	CAssert(m_utBlkBuf.poIn != NULL).App(MSG_MEMORY_ALLOC, _T("CBlockXform()")).Throw();
	m_utBlkBuf.poOut = new BYTE[m_utBlkBuf.dwBlkSiz];
	CAssert(m_utBlkBuf.poOut != NULL).App(MSG_MEMORY_ALLOC, _T("CBlockXform()")).Throw();
	m_utBlkBuf.iIn = m_utBlkBuf.iOut = 0;
}

CBlockXform::~CBlockXform() {
	delete m_utBlkBuf.poIn;
	delete m_utBlkBuf.poOut;
}
//
//	Generic encrypt/decrypt block-buffering, ensuring that the stream is treated block-wise, regardless
//	of how the data is fed in, and how space is made available out.
//
//	If any motion of data is possible, it will be done. Not necessarily *all* though, several calls
//	may be needed given any specific "starting" point.
//
void
CBlockXform::Xform() {
	DWORD i, j;
	// If we can/need to write anything from the block buffer...
	if (i = Min(m_utBlkBuf.iOut, utZstream.avail_out)) {
		CopyMemory(utZstream.next_out, &m_utBlkBuf.poOut[m_utBlkBuf.dwBlkSiz - m_utBlkBuf.iOut], i);
		m_utBlkBuf.iOut -= i;
		ConsumedInOut(0, i);
	}
	if (m_utBlkBuf.iOut) return;			// Partial output block left. Need more avail_out to continue.
	if (utZstream.avail_in == 0) return;	// Nothing more to do.

	// Check if we need to buffer a partial input block.
	// Mininum of avail_in and available space in block buffer (if dwBlkSiz it is empty),
	// thus the maximum value of this is dwBlkSiz, indicating that at least one complete
	// block is available for input.
	i = Min(utZstream.avail_in, (DWORD)(m_utBlkBuf.dwBlkSiz - m_utBlkBuf.iIn));
	// Calculate how many complete blocks can be processed from in to out, which
	// is the minimum of # of blocks in in-stream and # of blocks in out-stream.
	j = Min(utZstream.avail_in, utZstream.avail_out) / m_utBlkBuf.dwBlkSiz;	// # of blocks.

	// if less than a block is avail_in or there is data in in the block input buffer
	// OR
	// if there is at least a full block avail_in but there is not a full block avail_out.
	if (i < m_utBlkBuf.dwBlkSiz || // If this is FALSE, at least 1 block is availble from in-stream...
		j == 0) { // ... thus if this is TRUE, the problem is that wee need more out-stream space,
				  // but we can consume from in-stream to block buffer, which we do.
		// Copy data from instream to the buffer.
		CopyMemory(m_utBlkBuf.poIn + m_utBlkBuf.iIn, utZstream.next_in, i);
		m_utBlkBuf.iIn += i;
		ConsumedInOut(i, 0);
		// Check to see if we have a full block ready for encryption/decryption
		if (m_utBlkBuf.iIn == m_utBlkBuf.dwBlkSiz) {
			Xblock((TBlock *)m_utBlkBuf.poIn, (TBlock *)m_utBlkBuf.poOut);
			m_utBlkBuf.iOut = m_utBlkBuf.dwBlkSiz;
			m_utBlkBuf.iIn = 0;
		}
		return;								// Enough so far, we'll be back shortly at the top...
	} else {
		// This is a slight optimization - if one relaxes the first constraint above, all data
		// will go through that part of the code, one block at a time. But most data will be
		// in large chunks, so we do it in one call here, and have the loop deeper down instead
		// saving some overhead/block as well as bypassing the block-buffer.
		// No partial blocks remain, and we have at least one full block in the instream.
		Xblock((TBlock *)utZstream.next_in, (TBlock *)utZstream.next_out, j);
		ConsumedInOut(j * m_utBlkBuf.dwBlkSiz, j * m_utBlkBuf.dwBlkSiz);
	}
	// Happy so far. If there is a partial in-block left, next round will handle it.
}
// Constructor
CEncrypt::CEncrypt(TKey *putKey, TBlock *putIV, HWND hProgressWnd) : CBlockXform(hProgressWnd) {
	m_utAesCryptCBC.Init(putKey, CAes::eCBC, CAes::eEncrypt);
	m_utAesCryptCBC.SetIV(putIV);
	m_bLast = FALSE;
}

/*virtual*/ void
CEncrypt::Xblock(TBlock *putSrc, TBlock *putDst, DWORD dwN) {
	m_utAesCryptCBC.Xblock(putSrc, putDst, dwN);
}
//
//	When we get here avail_in is zero, so all in-stream data is consumed.
//	If the last block was filled, we known that "Xform" will always move
//	that to the out-buffer in the same call as it is consumed.
//
//	Thus, after emptying the out-buffer, we either have data to be padded
//	in in-buffer or an empty in-buffer that needs to be filled with all
//	padding.
//
//	This function may be called multiple times. Return TRUE if there
//	is more to do, and you would like to be called again...
//
int
CEncrypt::Finish() {
	Xform();							// Flush possibly full outbuffer block.
	if (m_utBlkBuf.iOut) return TRUE;	// Need more outbuffer before proceeding

	if (!m_bLast) {
		// Fill the in-buffer with padding according to RFC 1423 adapted to 16-byte blocks
		for (DWORD i = m_utBlkBuf.iIn; i < m_utBlkBuf.dwBlkSiz; i++) {
			m_utBlkBuf.poIn[i] = (BYTE)(m_utBlkBuf.dwBlkSiz - m_utBlkBuf.iIn);
		}
		// Encrypt the block, and make it ready for output.
		Xblock((TBlock *)m_utBlkBuf.poIn, (TBlock *)m_utBlkBuf.poOut);
		m_utBlkBuf.iIn = 0;
		m_utBlkBuf.iOut = m_utBlkBuf.dwBlkSiz;
		m_bLast = TRUE;
		return TRUE;					// Come back for more!
	}
	// No data in out-block-buf, Have made the padding block -> done!
	return FALSE;
}
//
//	Knowing the padding algorithm, we calculate what the total size of the data produced will be.
//	Standard padding will pad up to even m_utBlkBuf.dwBlkSiz-byte boundary, when data is even m_utBlkBuf.dwBlkSiz-byte, an extra
//	block of only padding will be produced.
//
//	The padding scheme is from RFC 1423 adapted to 16-byte blocks
//
BYTE
CEncrypt::GetPadSize(QWORD qwIn) {
	return (BYTE)(m_utBlkBuf.dwBlkSiz - qwIn % m_utBlkBuf.dwBlkSiz);
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CEncrypt::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_ENCRYPT).GetMsg());
    }
}
// Constructor
CDecrypt::CDecrypt(TKey *putKey, TBlock *putIV, HWND hProgressWnd) : CBlockXform(hProgressWnd) {
	m_utAesCryptCBC.Init(putKey, CAes::eCBC, CAes::eDecrypt);
	m_utAesCryptCBC.SetIV(putIV);
}

/*virtual*/ void
CDecrypt::Xblock(TBlock *putSrc, TBlock *putDst, DWORD dwN) {
	m_utAesCryptCBC.Xblock(putSrc, putDst, dwN);
}
//
//	Return TRUE while more data to flush! We know here there is no more data to read.
//	We also know that as the input is padded, we have a complete block waiting
//	for output, including padding. In fact, the entire block may well be padding!
//
//	The data about the padding length is in the block.
//	The padding scheme is from RFC 1423 adapted to 16-byte blocks
//
//	Two general cases exist: A full block of padding, or a mixed block with both padding
//	and data. Regardless, we known that the current block, which will always be in the
//	outbuffer, possibly partially consumed, will have padding.
//
int
CDecrypt::Finish() {
	Xform();							// Flush (possibly full) outbuffer block.

	// At this point, we should only have padding left as output data, assuming
	// there was enough space in the output buffer to contain it.
	unsigned int iPad = m_utBlkBuf.poOut[m_utBlkBuf.dwBlkSiz - 1];
	// If we have more than pad left in the outbuffer...
	if (m_utBlkBuf.iOut > iPad) {
		return TRUE;	// Need more outbuffer before proceeding
	}
	// Before finally ending - check that the padding is correct.
	for (unsigned int i = m_utBlkBuf.dwBlkSiz - iPad; i < m_utBlkBuf.dwBlkSiz; i++) {
		CAssert(m_utBlkBuf.poOut[i] == iPad).App(MSG_PAD_ERROR).Throw();
	}
	return FALSE;
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CDecrypt::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_DECRYPT).GetMsg());
    }
}
//
//	Calculate the RFC2104 HMAC
//
CHmac::CHmac(TKey *pDataEncKey, HWND hProgressWnd) : CXform(hProgressWnd) {
	m_putContext = new SHA1_CTX;
	m_putHMAC = new THash;

	// Generate the sub key
	m_utHMACKey.Set(pDataEncKey, CSubKey::eHMAC);
}
//
//	Destroy the temporary values. Important that they are on the heap.
//
CHmac::~CHmac() {
	if (m_putContext != NULL) delete m_putContext;
	if (m_putHMAC != NULL) delete m_putHMAC;
}
//
//	Do the initial inner padding, xor and hashing of that.
//
/*virtual*/ QWORD
CHmac::Init(CFileIO& utInFile) {
	SHA1Init(m_putContext);

	// K xor ipad
	XorPad(0x36);
	SHA1Update(m_putContext, (BYTE *)m_putHMAC, sizeof *m_putHMAC);
    return utInFile.m_qwFileSize - utInFile.m_qwFileOffset;
}
//
//	Hash all data in the inner 'loop'.
//
/*virtual*/ void
CHmac::Xform() {
	SHA1Update(m_putContext, utZstream.next_in, utZstream.avail_in);
	ConsumedInOut(utZstream.avail_in, 0);
}
//
//	Take the MAC key, xor with the outer value and pad,
//	append the previous hash, and produce the final
//	HMAC-SHA1 result.
//
/*virtual*/ int
CHmac::Finish() {
	THash *putHash = new THash;
	SHA1Final((BYTE *)putHash, m_putContext);

	// K xor opad
	XorPad(0x5c);
	// Now do the outer hash.
	SHA1Init(m_putContext);
	SHA1Update(m_putContext, (BYTE *)m_putHMAC, sizeof *m_putHMAC);
	SHA1Update(m_putContext, (BYTE *)putHash, sizeof *putHash);
	SHA1Final((BYTE *)m_putHMAC, m_putContext);
	delete putHash;
	return FALSE;
}
//
//	Return the result. The Hmac() call may truncate.
//
THmac *
CHmac::GetHMAC() {
	return m_putHMAC->Hmac();
}
//
//	Helper for the inner and outer padding and Xor
//
void CHmac::XorPad(BYTE oPad) {
	for  (int i=0; i < sizeof *m_putHMAC; i++) {
		((BYTE *)m_putHMAC)[i] = oPad;
		if (i < sizeof *m_utHMACKey.Get()) {
			((BYTE *)m_putHMAC)[i] ^= ((BYTE *)m_utHMACKey.Get())[i];
		}
	}
}
//
// Name the operation for the progress window
//
/*virtual*/ void
CHmac::StartProgress() {
    CXform::StartProgress();
    if (ProgressWnd() != NULL) {
        // Need to use SendMessage to cross process boundary
        SendMessage(GetDlgItem(GetParent(ProgressWnd()), IDS_OPERATION), WM_SETTEXT, 0, (LPARAM)(LPTSTR)CMessage().AppMsg(INF_OPNAME_HMAC).GetMsg());
    }
}
